/*********************************************************************

  png.c

  PNG reading functions.

  07/15/1998 Created by Mathis Rosenhauer
  10/02/1998 Code clean up and abstraction by Mike Balfour
             and Mathis Rosenhauer
  10/15/1998 Image filtering. MLR
  11/09/1998 Bit depths 1-8 MLR
  11/10/1998 Some additional PNG chunks recognized MLR
  05/14/1999 Color type 2 and PNG save functions added
  05/15/1999 Handle RGB555 while saving, use osd_fxxx
             functions for writing MSH

  TODO : Fully comply with the "Recommendations for Decoders"
         of the W3C

*********************************************************************/

#include <math.h>
#include <zlib.h>
#include "main.h"
#include "png.h"

/* convert_uint is here so we don't have to deal with byte-ordering issues */
static UINT32 convert_from_network_order (UINT8 *v)
{
	UINT32 i;

	i = (v[0]<<24) | (v[1]<<16) | (v[2]<<8) | (v[3]);
	return i;
}

int png_unfilter(struct png_info *p)
{
	int i, j, bpp, filter;
	INT32 prediction, pA, pB, pC, dA, dB, dC;
	UINT8 *src, *dst;
   SOUND_UPDATE_INIT

	if((p->image = (UINT8 *)malloc (p->height*p->rowbytes))==NULL)
	{
		free (p->fimage);
		return 0;
	}

	src = p->fimage;
	dst = p->image;
	bpp = p->bpp;

	for (i=0; i<p->height; i++)
	{
      SOUND_UPDATE_TRIP(200)
      filter = *src++;
		if (!filter)
		{
			memcpy (dst, src, p->rowbytes);
			src += p->rowbytes;
			dst += p->rowbytes;
		}
		else
			for (j=0; j<p->rowbytes; j++)
			{
				pA = (j<bpp) ? 0: *(dst - bpp);
				pB = (i<1) ? 0: *(dst - p->rowbytes);
				pC = ((j<bpp)||(i<1)) ? 0: *(dst - p->rowbytes - bpp);

				switch (filter)
				{
				case PNG_PF_Sub:
					prediction = pA;
					break;
				case PNG_PF_Up:
					prediction = pB;
					break;
				case PNG_PF_Average:
					prediction = ((pA + pB) / 2);
					break;
				case PNG_PF_Paeth:
					prediction = pA + pB - pC;
					dA = abs(prediction - pA);
					dB = abs(prediction - pB);
					dC = abs(prediction - pC);
					if (dA <= dB && dA <= dC) prediction = pA;
					else if (dB <= dC) prediction = pB;
					else prediction = pC;
					break;
				default:
					prediction = 0;
				}
				*dst = 0xff & (*src + prediction);
				dst++; src++;
			}
	}

	free (p->fimage);
	return 1;
}
int png_verify_signature (FILE *fp)
{
	INT8 signature[8];

	if (fread (signature,1, 8,fp) != 8)
		return 0;

	if (memcmp(signature, PNG_Signature, 8))
		return 0;
	return 1;
}

int png_inflate_image (struct png_info *p)
{
	unsigned long fbuff_size;

    fbuff_size = p->height * (p->rowbytes + 1);

	if((p->fimage = (UINT8 *)malloc (fbuff_size))==NULL)
	{
      free (p->zimage);
		return 0;
	}

   if (uncompress(p->fimage, &fbuff_size, p->zimage, p->zlength) != Z_OK)
	{
		return 0;
	}

	free (p->zimage);
	return 1;
}

int png_read_file(FILE *fp, struct png_info *p)
{
	/* translates color_type to bytes per pixel */
	const int samples[] = {1, 0, 3, 1, 2, 0, 4};
	int i;
   SOUND_UPDATE_INIT

	UINT32 chunk_length,chunk_type=0, chunk_crc,crc;
	UINT32 tot_idat=0, num_idat=0, l_idat[256];
	UINT8 *chunk_data, *zimage, *idat[256], *temp;
	UINT8 str_chunk_type[5], v[4];

	p->num_palette = 0;
	p->num_trans = 0;
	p->trans = NULL;
	p->palette = NULL;

	if (png_verify_signature(fp)==0)
		return 0;

	while (chunk_type != PNG_CN_IEND)
	{
      SOUND_UPDATE_TRIP(200)
      fread(v, 1, 4,fp);
		chunk_length=convert_from_network_order(v);

		fread(str_chunk_type,1, 4,fp);

		str_chunk_type[4]=0; /* terminate string */

		crc=crc32(0,str_chunk_type, 4);
		chunk_type = convert_from_network_order(str_chunk_type);

		if (chunk_length)
		{
			if ((chunk_data = (UINT8 *)malloc(chunk_length+1))==NULL)
			{
				return 0;
			}
			if (fread (chunk_data, 1, chunk_length,fp) != chunk_length)
			{
				free(chunk_data);
				return 0;
			}

			crc=crc32(crc,chunk_data, chunk_length);
		}
		else
			chunk_data = NULL;

		fread(v,1, 4,fp);
		chunk_crc=convert_from_network_order(v);

		if (crc != chunk_crc)
		{
			return 0;
		}


		switch (chunk_type)
		{
		case PNG_CN_IHDR:
			p->width = convert_from_network_order(chunk_data);
			p->height = convert_from_network_order(chunk_data+4);
			p->bit_depth = *(chunk_data+8);
			p->color_type = *(chunk_data+9);
			p->compression_method = *(chunk_data+10);
			p->filter_method = *(chunk_data+11);
			p->interlace_method = *(chunk_data+12);
			free (chunk_data);

			break;

		case PNG_CN_PLTE:
			p->num_palette=chunk_length/3;
			p->palette=chunk_data;
			break;

		case PNG_CN_tRNS:
			p->num_trans=chunk_length;
			p->trans=chunk_data;
			break;

		case PNG_CN_IDAT:
			idat[num_idat]=chunk_data;
			l_idat[num_idat]=chunk_length;
			num_idat++;
			tot_idat += chunk_length;
			break;

		case PNG_CN_tEXt:
			free(chunk_data);
			break;

		case PNG_CN_tIME:

			free(chunk_data);
			break;

		case PNG_CN_gAMA:
			p->source_gamma  = convert_from_network_order(chunk_data)/100000.0;
			free(chunk_data);
			break;

		case PNG_CN_pHYs:
			p->xres = convert_from_network_order(chunk_data);
			p->yres = convert_from_network_order(chunk_data+4);
			p->resolution_unit = *(chunk_data+8);
			free(chunk_data);
			break;

		case PNG_CN_IEND:
			break;

		default:
			if (chunk_data)
				free(chunk_data);
			break;
		}
	}
	if ((zimage = (UINT8 *)malloc(tot_idat))==NULL)
	{
		return 0;
	}

	p->zlength = tot_idat;

	/* combine idat chunks to compressed image data */
	temp = zimage;
	for (i=0; i<num_idat; i++)
	{
		memcpy (temp, idat[i], l_idat[i]);
		free (idat[i]);
		temp += l_idat[i];
	}
	p->zimage=zimage;
	p->bpp = (samples[p->color_type] * p->bit_depth) / 8;
	p->rowbytes = ceil((p->width * p->bit_depth * samples[p->color_type]) / 8.0);

	if (png_inflate_image(p)==0)
		return 0;

	if(png_unfilter (p)==0)
		return 0;

	return 1;
}

/*  Expands a p->image from p->bit_depth to 8 bit */
int png_expand_buffer_8bit (struct png_info *p)
{
	int i,j, k;
	UINT8 *pinp, *poutp, *outbuf;
   SOUND_UPDATE_INIT

	if (p->bit_depth < 8)
	{
		if ((outbuf = (UINT8 *)malloc(p->width*p->height))==NULL)
		{
			return 0;
		}

		pinp = p->image;
		poutp = outbuf;

		for (i = 0; i < p->height; i++)
		{
         SOUND_UPDATE_TRIP(200)
         for(j = 0; j < p->width / ( 8 / p->bit_depth); j++)
			{
				for (k = 8 / p->bit_depth-1; k >= 0; k--)
					*poutp++ = (*pinp >> k * p->bit_depth) & (0xff >> (8 - p->bit_depth));
				pinp++;
			}
			if (p->width % (8 / p->bit_depth))
			{
				for (k = p->width % (8 / p->bit_depth)-1; k >= 0; k--)
					*poutp++ = (*pinp >> k * p->bit_depth) & (0xff >> (8 - p->bit_depth));
				pinp++;
			}
		}
		free (p->image);
		p->image = outbuf;
	}

	return 1;
}

void png_delete_unused_colors (struct png_info *p)
{
	int i, tab[256], pen=0, trns=0;
	UINT8 ptemp[3*256], ttemp[256];

	memset (tab, 0, 256*sizeof(int));
	memcpy (ptemp, p->palette, 3*p->num_palette);
	memcpy (ttemp, p->trans, p->num_trans);

	/* check which colors are actually used */
	for (i = 0; i < p->height*p->width; i++)
		tab[p->image[i]]++;

	/* shrink palette and transparency */
	for (i = 0; i < p->num_palette; i++)
		if (tab[i])
		{
			p->palette[3*pen+0]=ptemp[3*i+0];
			p->palette[3*pen+1]=ptemp[3*i+1];
			p->palette[3*pen+2]=ptemp[3*i+2];
			if (i < p->num_trans)
			{
				p->trans[pen] = ttemp[i];
				trns++;
			}
			tab[i] = pen++;
		}

	/* remap colors */
	for (i = 0; i < p->height*p->width; i++)
		p->image[i]=tab[p->image[i]];


	p->num_palette = pen;
	p->num_trans = trns;
}
int loadpng(char *filename,PPCXSTRUCT pcx)
{
	FILE *fp;
	struct png_info p;
	pBYTE outline;
   pBYTE pfInBuffer;
   pBYTE pfImage;
   pBYTE pfWork;
   int h,w,y;
   int xcomp,ycomp,perc;
   int xcnt,ycnt,xerror,yerror;
   int saveout,outpos;
   int bAlwaysX=FALSE;
   int xx,nWork;
   int XSize,YSize;
   SOUND_UPDATE_INIT


	fp=fopen(filename,"rb");
	if(!fp)
		return 0;

	if (!png_read_file(fp, &p))
	{
		fclose (fp);
		return 0;
	}
	fclose (fp);

	if (p.bit_depth > 8)
	{
      loginfo(1,"Unsupported PNG bit depth %i (8 bit max.)", p.bit_depth);
      return 0;
	}

	if (p.color_type != 3)
	{
      loginfo(1,"Unsupported PNG color type %i (has to be 3)", p.color_type);
      return 0;
	}

	if (p.interlace_method != 0)
	{
      loginfo(1,"PNG Interlace unsupported");
		return 0;
	}

	/* Convert to 8 bit */
	png_expand_buffer_8bit (&p);
	png_delete_unused_colors (&p);


   if(!noRotatePics)
   {
      switch(ScreenOrient.Orientation)
      {
         case ORIENT_NORMAL:
            XSize=ScreenOrient.xpixel;
            YSize=ScreenOrient.ypixel;
            break;
         default:
            XSize=ScreenOrient.ypixel;
            YSize=ScreenOrient.xpixel;
      }
   }
   else
   {
      XSize=ScreenOrient.xpixel;
      YSize=ScreenOrient.ypixel;
   }


//get the dimensions
	h=p.height;
	w=p.width;

//find the biggest dimension
	outpos=0;

   xcomp=w-XSize;
   ycomp=h-YSize;
	if(xcomp>0||ycomp>0)
	{
		if(xcomp>ycomp)
//scale x to fit
         perc=(int)((XSize*100)/(long)w);
		else
//scale y to fit
         perc=(int)((YSize*100)/(long)h);
//make them both fit
		while(TRUE)
		{
         SOUND_UPDATE_TRIP(200)
			xcomp=(int)((long)((long)perc*(long)w)/100L);
			ycomp=(int)((long)((long)perc*(long)h)/100L);
         if(xcomp>XSize||ycomp>YSize)
			{
				perc--;
				if(perc==0)
					break;
			}
			else
				break;
		}
	}
	else
    {
        if(!noScaleToFit)
        {
            if(xcomp<ycomp)
//scale x to fit
               perc=(int)((XSize*100)/(long)w);
            else
//scale y to fit
               perc=(int)((YSize*100)/(long)h);
//make them both fit
            while(TRUE)
            {
                xcomp=(int)((long)((long)perc*(long)w)/100L);
                ycomp=(int)((long)((long)perc*(long)h)/100L);
                if(xcomp>XSize||ycomp>YSize)
                {
                    perc--;
                    if(perc==0)
                        break;
                }
                else
                    break;
            }
			if(perc<100)
				perc=100;
        }
        else
            perc=100;
    }


   pfImage=pcx->image;
	clearimage(pfImage,&p);


//Loop for entire length of bitmap buffer:
	xerror=yerror=0;
	xcnt=ycnt=0;

//center the x
	if(perc==100)
	{
      outpos=((XSize-w)>>1)+(((YSize-h)>>1)*XSize);
      if(noScaleToFit)
      {
         switch(SSPosition)
         {
            case SSPOS_OFFSET+1:
               outpos=XSize-w;
               break;
            case SSPOS_OFFSET+2:
               outpos=XSize-w+((YSize-h)*XSize);
               break;
            case SSPOS_OFFSET+3:
               outpos=XSize-w+(((YSize-h)>>1)*XSize);
               break;
         }
      }
		saveout=outpos;
	}
	else
	{
      if(w>XSize||(w<XSize&&!noScaleToFit))
		{
			nWork=(int)(((long)w*(long)perc)/100L);
		}
		else
		{
			nWork=w;
			bAlwaysX=TRUE;

		}
      outpos=(XSize-nWork);
		outpos>>=1;
//centre y
      nWork=(YSize-((int)(((long)h*(long)perc)/100L)))>>1;
      outpos+=(nWork*XSize);
      saveout=outpos;
   }

	for(y=0;y<p.height;y++)
	{
      SOUND_UPDATE_TRIP(200)
		outline=&p.image[w*y];
		if(perc==100)
		{
			for(xx=0;xx<w;xx++)
				pfImage[outpos++]=outline[xx];
            outpos=saveout+XSize;
			saveout=outpos;
		}
		else
		{
			yerror+=perc;
         if(yerror>100)
         {
            while(yerror>100)
            {
               if(bAlwaysX)
               {
                  for(xx=0;xx<w;xx++)
                     pfImage[outpos++]=outline[xx];
               }
               else
               {
                  xerror=0;
                  for(xx=0;xx<w;xx++)
                  {
                     xerror+=perc;
                     while(xerror>100)
                     {
                        pfImage[outpos++]=outline[xx];
                        xerror-=100;
                     }
               }
            }
            outpos=saveout+XSize;
            saveout=outpos;
            yerror-=100;
         }
      }
      else
      {
         if(noScaleToFit)
            memset(outline,0,w);
      }
   }
   if(outpos>=ScreenOrient.memsize)
      break;
   }

	for(y =0;y< p.num_palette && y<256;y++)
	{
		pcx->palette[y*3]=p.palette[y*3]>>2;
		pcx->palette[(y*3)+1]=(p.palette[(y*3)+1]>>2);
		pcx->palette[(y*3)+2]=(p.palette[(y*3)+2]>>2);
	}

    if(!noRotatePics)
    {
        pfWork=myalloc(ScreenOrient.memsize);
        switch(ScreenOrient.Orientation)
        {
            case ORIENT_ROTATED_LEFT:
                for(h=0;h<ScreenOrient.ypixel;h++)
                    for(w=0;w<ScreenOrient.xpixel;w++)
                        pfWork[w+(((ScreenOrient.ypixel-1)-h)*ScreenOrient.xpixel)]=pfImage[(w*XSize)+h];
                memcpy(pfImage,pfWork,ScreenOrient.memsize);
                break;
            case ORIENT_ROTATED_RIGHT:
                for(h=0;h<ScreenOrient.ypixel;h++)
                    for(w=0;w<ScreenOrient.xpixel;w++)
                        pfWork[((ScreenOrient.xpixel-1)-w)+(h*ScreenOrient.xpixel)]=pfImage[(w*XSize)+h];
                memcpy(pfImage,pfWork,ScreenOrient.memsize);

      }

      myfree(pfWork);
   }

   myfree(p.image);
   myfree(p.palette);
	return 1;
}
/* End of source */
void clearimage(pBYTE image,struct png_info *p)
{
	BYTE black=0;
	int n,found=0;
//find black
	for(n=0;n<p->num_palette;n++)
	{
		if(!p->palette[n*3]&&!p->palette[(n*3)+1]&&!p->palette[(n*3)+2])
		{
			black=(BYTE)n;
			found=1;
		}
	}
	if(!found)
	{
		p->palette[p->num_palette*3]=0;
		p->palette[(p->num_palette*3)+1]=1;
		p->palette[(p->num_palette*3)+2]=2;
		black=(BYTE)p->num_palette;
	}
	memset(image,black,ScreenOrient.memsize);
}
